
import logging
import collections

from prophetess.plugin import Loader
from prophetess_netbox.client import NetboxClient


log = logging.getLogger('prophetess.plugins.netbox.loader')


class NetboxLoader(Loader):
    required_config = (
        'host',
        'api_key',
        'endpoint',
        'model',
        'pk',
    )

    def __init__(self, **kwargs):
        """ NetboxLoader init """
        super().__init__(**kwargs)

        self.update_method = self.config.get('update_method', 'update')
        self.client = NetboxClient(host=self.config.get('host'), api_key=self.config.get('api_key'))

    def sanitize_config(self, config):
        """ Overload Loader.sanitize_config to add additional conditioning """
        config = super().sanitize_config(config)

        for k in ('model', 'endpoint'):
            config[k] = config[k].lower()

        if not isinstance(config['pk'], list):
            config['pk'] = [config['pk']]

        return config

    async def parse_fk(self, record):
        extracts = self.config.get('fk')
        if not extracts or not isinstance(extracts, collections.Mapping):
            return record

        for key, rules in extracts.items():
            if key not in record:
                log.debug('Skipping FK lookup "{}". Not found in record'.format(key))
                continue

            r = await self.client.entity(
                endpoint=rules.get('endpoint'),
                model=rules.get('model'),
                params=self.build_params(rules.get('pk', []), record)
            )

            if not r:
                log.debug('FK lookup for {} ({}) failed, no record found'.format(key, record.get(key)))
                record[key] = None
                continue

            record[key] = r.id

        return record

    def build_params(self, config, record):
        output = {}
        for item in config:
            if isinstance(item, str):
                output[item] = record.get(item)
            elif isinstance(item, collections.Mapping):
                for k, tpl in item.items():
                    output[k] = tpl.format(**record)

        return output

    def diff_records(self, cur_record, new_record):

        changed = {}

        for k, v in new_record.items():
            if getattr(cur_record, k) and k in self.config.get('fk', {}):
                if getattr(cur_record, k).id != v:
                    changed[k] = v
                continue

            if getattr(cur_record, k) != v:
                changed[k] = v

        return changed

    async def run(self, record):
        """ Overload Loader.run to execute netbox loading of a record """

        er = await self.client.entity(
            endpoint=self.config.get('endpoint'),
            model=self.config.get('model'),
            params=self.build_params(self.config.get('pk'), record)
        )

        record = await self.parse_fk(record)

        payload = {
            'data': record
        }

        method = 'create'
        if er:
            method = self.update_method
            payload['id'] = er.id

        if method == 'partial_update':
            changed_record = self.diff_records(er, record)
            if not changed_record:
                log.debug('Skipping {} as no data has changed'.format(record))
                return

            payload['data'] = changed_record

        func = self.client.build_model(self.config.get('endpoint'), self.config.get('model'), method)

        log.debug('Running with modified payload: {}'.format(payload))
        return await func(**payload)

    async def close(self):
        await self.client.close()



import asynctest
from unittest.mock import MagicMock


class AIONetboxResponseMock():
    pass


class AIONetboxMagicMock(MagicMock):

    request = asynctest.CoroutineMock(return_value=AIONetboxResponseMock())
    close = asynctest.CoroutineMock()


class AIONetboxMock():

    _attrs = {}

    request = asynctest.CoroutineMock(return_value=AIONetboxResponseMock())
    close = asynctest.CoroutineMock()

    def __init__(*args, **kwargs):
        pass

    def __setattr__(self, attr, value):
        self._attrs[attr] = value

    def __getattr__(self, attr):

        ret = self._attrs.get(attr, MagicMock())

        if isinstance(ret, Exception):
            raise ret

        return ret




import pytest
import asynctest

from unittest.mock import patch

from prophetess_netbox.loader import NetboxLoader
from .fixtures import AIONetboxMagicMock, AIONetboxResponseMock


@patch('prophetess_netbox.loader.NetboxClient')
def test_NetboxLoader(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'pk': [],
    }

    nbl = NetboxLoader(id='nbloader', config=config)

    mnbc.assert_called_with(
        host='http://testing',
        api_key='12test'
    )

    assert nbl.update_method == 'update'


@patch('prophetess_netbox.loader.NetboxClient')
def test_NetboxLoader_sanitize_config(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'DCIM',
        'model': 'Sites',
        'pk': 'slug',
    }

    nbl = NetboxLoader(id='nbloader', config=config)

    sanitized = nbl.sanitize_config(config)

    assert 'dcim' == sanitized['endpoint']
    assert 'sites' == sanitized['model']
    assert ['slug'] == sanitized['pk']


@pytest.mark.asyncio
@patch('prophetess_netbox.loader.NetboxClient')
async def test_NetboxLoader_parse_fk(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'pk': [],
        'fk': {
            'tenant': {
                'endpoint': 'tenant',
                'model': 'tenants',
                'pk': [
                    {
                        'slug': '{tenant}',
                    },
                ],
            },
        },
    }

    nbl = NetboxLoader(id='nbloader', config=config)
    mnbc.return_value.entity = asynctest.CoroutineMock()
    mnbc.return_value.entity.return_value = AIONetboxResponseMock()
    mnbc.return_value.entity.return_value.id = 1

    ret = await nbl.parse_fk({'id': 'yay', 'tenant': 'fk-lookup-plz'})

    assert {'id': 'yay', 'tenant': 1} == ret

    mnbc.return_value.entity.assert_called_with(
        endpoint='tenant',
        model='tenants',
        params={'slug': 'fk-lookup-plz'},
    )


@pytest.mark.asyncio
@patch('prophetess_netbox.loader.NetboxClient')
async def test_NetboxLoader_parse_fk_missing(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'pk': [],
    }

    nbl = NetboxLoader(id='nbloader', config=config)
    ret = await nbl.parse_fk({'foo': 'bar'})

    assert {'foo': 'bar'} == ret


@pytest.mark.asyncio
@patch('prophetess_netbox.loader.NetboxClient')
async def test_NetboxLoader_parse_fk_failed(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'pk': [],
        'fk': {
            'tenant': {
                'endpoint': 'tenant',
                'model': 'tenants',
                'pk': [
                    {
                        'slug': '{tenant}',
                    },
                ],
            },
        },
    }

    nbl = NetboxLoader(id='nbloader', config=config)
    mnbc.return_value.entity = asynctest.CoroutineMock()
    mnbc.return_value.entity.return_value = None

    ret = await nbl.parse_fk({'id': 'yay', 'tenant': 'fk-lookup-plz'})

    assert {'id': 'yay', 'tenant': None} == ret


@pytest.mark.asyncio
@patch('prophetess_netbox.loader.NetboxClient')
async def test_NetboxLoader_parse_fk_skip(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'pk': [],
        'fk': {
            'tenant': {
                'endpoint': 'tenant',
                'model': 'tenants',
                'pk': [
                    {
                        'slug': '{tenant}',
                    },
                ],
            },
        },
    }

    nbl = NetboxLoader(id='nbloader', config=config)

    ret = await nbl.parse_fk({'id': 'yay'})

    assert {'id': 'yay'} == ret


@patch('prophetess_netbox.loader.NetboxClient')
def test_NetboxLoader_build_params(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'pk': [
            'slug',
            {'foo': '{data}'}
        ],
    }

    record = {
        'slug': 'hello-dolly',
        'data': 'this is foo',
        'not-pk': 'just ignore',
    }

    nbl = NetboxLoader(id='nbloader', config=config)
    params = nbl.build_params(config.get('pk'), record)

    assert {'slug': 'hello-dolly', 'foo': 'this is foo'} == params


@patch('prophetess_netbox.loader.NetboxClient')
def test_NetboxLoader_diff_records(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'pk': [],
        'fk': {
            'tenant': {
                'endpoint': 'tenant',
                'model': 'tenants',
                'pk': [
                    {
                        'slug': '{tenant}',
                    },
                ],
            },
        },
    }

    cur_record = AIONetboxResponseMock()

    cur_record.id = 1
    cur_record.slug = 'hello'
    cur_record.tenant = AIONetboxResponseMock()
    cur_record.tenant.id = 2
    cur_record.tenant.name = 'hey'

    new_record = {
        'id': 1,
        'slug': 'hello',
        'tenant': 2,
    }

    nbl = NetboxLoader(id='nbloader', config=config)
    diff = nbl.diff_records(cur_record, new_record)

    assert {} == diff


@patch('prophetess_netbox.loader.NetboxClient')
def test_NetboxLoader_diff_records_changed(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'pk': [],
        'fk': {
            'tenant': {
                'endpoint': 'tenant',
                'model': 'tenants',
                'pk': [
                    {
                        'slug': '{tenant}',
                    },
                ],
            },
        },
    }

    cur_record = AIONetboxResponseMock()

    cur_record.id = 1
    cur_record.slug = 'hello'
    cur_record.tenant = AIONetboxResponseMock()
    cur_record.tenant.id = 2
    cur_record.tenant.name = 'hey'

    new_record = {
        'id': 1,
        'slug': 'goodbye',
        'tenant': 3,
    }

    nbl = NetboxLoader(id='nbloader', config=config)
    diff = nbl.diff_records(cur_record, new_record)

    assert {'slug': 'goodbye', 'tenant': 3} == diff


@pytest.mark.asyncio
@patch('prophetess_netbox.loader.NetboxClient')
async def test_NetboxLoader_run_create(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'pk': [],
    }

    record = {
        'id': 1,
        'slug': 'goodbye',
        'tenant': 3,
    }

    nbl = NetboxLoader(id='nbloader', config=config)
    mnbc.return_value.entity = asynctest.CoroutineMock()
    mnbc.return_value.entity.return_value = None
    mnbc.return_value.build_model.return_value = asynctest.CoroutineMock()

    ret = await nbl.run(record)

    assert mnbc.return_value.build_model.return_value.return_value == ret

    mnbc.return_value.build_model.assert_called_with('dcim', 'sites', 'create')
    mnbc.return_value.build_model.return_value.assert_called_with(data=record)


@pytest.mark.asyncio
@patch('prophetess_netbox.loader.NetboxClient')
async def test_NetboxLoader_run_update(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'update_method': 'partial_update',
        'pk': [],
    }

    record = {
        'id': 1,
        'slug': 'goodbye',
    }

    nbl = NetboxLoader(id='nbloader', config=config)
    mnbc.return_value.entity = asynctest.CoroutineMock()
    mnbc.return_value.entity.return_value = AIONetboxResponseMock()
    mnbc.return_value.entity.return_value.id = 42
    mnbc.return_value.entity.return_value.slug = 'updateme'
    mnbc.return_value.build_model.return_value = asynctest.CoroutineMock()

    await nbl.run(record)

    mnbc.return_value.build_model.assert_called_with('dcim', 'sites', 'partial_update')
    mnbc.return_value.build_model.return_value.assert_called_with(id=42, data=record)


@pytest.mark.asyncio
@patch('prophetess_netbox.loader.NetboxClient')
async def test_NetboxLoader_run_update_skip(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'update_method': 'partial_update',
        'pk': [],
    }

    record = {
        'id': 1,
        'slug': 'goodbye',
    }

    nbl = NetboxLoader(id='nbloader', config=config)
    mnbc.return_value.entity = asynctest.CoroutineMock()
    mnbc.return_value.entity.return_value = AIONetboxResponseMock()
    mnbc.return_value.entity.return_value.id = 1
    mnbc.return_value.entity.return_value.slug = 'goodbye'

    assert await nbl.run(record) is None


@pytest.mark.asyncio
@patch('prophetess_netbox.loader.NetboxClient', new_callable=AIONetboxMagicMock)
async def test_NetboxLoader_close(mnbc):

    config = {
        'host': 'http://testing',
        'api_key': '12test',
        'endpoint': 'dcim',
        'model': 'sites',
        'pk': [],
    }

    nbl = NetboxLoader(id='nbloader', config=config)
    await nbl.close()

    mnbc.return_value.close.assert_called()




import pytest
import asynctest

from unittest.mock import patch

from prophetess_netbox.client import NetboxClient
from prophetess_netbox.exceptions import InvalidPKConfig, InvalidNetboxEndpoint, InvalidNetboxOperation
from .fixtures import AIONetboxMock, AIONetboxMagicMock, AIONetboxResponseMock


@patch('prophetess_netbox.client.AIONetbox')
def test_NetboxClient(maionb):
    NetboxClient(host='http://test', api_key='key')

    maionb.from_openapi.assert_called_with(
        url='http://test',
        api_key='key'
    )


@pytest.mark.asyncio
@patch('prophetess_netbox.client.AIONetbox', new_callable=AIONetboxMagicMock)
async def test_NetboxClient_close(maionb):
    nb = NetboxClient(host='http://test', api_key='key')

    await nb.close()

    maionb.close.assert_called_once()


@patch('prophetess_netbox.client.AIONetbox', new_callable=AIONetboxMagicMock)
def test_NetboxClient_build_model(maionb):
    nb = NetboxClient(host='http://test', api_key='key')
    client = maionb.from_openapi.return_value
    model = nb.build_model('api', 'test', 'get')

    assert model == client.api.api_test_get


@patch('prophetess_netbox.client.AIONetbox', new_callable=AIONetboxMagicMock)
def test_NetboxClient_build_model_invalid_endpoint(maionb):
    maionb.from_openapi = AIONetboxMock
    nb = NetboxClient(host='http://test', api_key='key')
    nb.client.api = AttributeError()

    with pytest.raises(InvalidNetboxEndpoint):
        nb.build_model('api', 'test', 'get')


@patch('prophetess_netbox.client.AIONetbox', new_callable=AIONetboxMock)
def test_NetboxClient_build_model_invalid_operation(maionb):
    maionb.from_openapi = AIONetboxMagicMock

    nb = NetboxClient(host='http://test', api_key='key')

    nb.client.api = AIONetboxMock()
    nb.client.api.api_bad_get = AttributeError()
    with pytest.raises(InvalidNetboxOperation):
        nb.build_model('api', 'bad', 'get')


@pytest.mark.asyncio
@patch('prophetess_netbox.client.AIONetbox')
async def test_NetboxClient_fetch(__aionb):

    with patch.object(NetboxClient, 'build_model') as mbm:
        mbm.return_value = asynctest.CoroutineMock()

        nb = NetboxClient(host='http://test', api_key='key')
        await nb.fetch(
            endpoint='test',
            model='testing',
            params={
                'foo': 'bar',
                'baz': 'buzz',
            }
        )

        mbm.assert_called_with('test', 'testing', 'list')
        mbm.return_value.assert_called_with(foo='bar', baz='buzz')


@pytest.mark.asyncio
@patch('prophetess_netbox.client.AIONetbox')
async def test_NetboxClient_fetch_invalid(__aionb):

    with patch.object(NetboxClient, 'build_model') as mbm:
        mbm.return_value.side_effect = [ValueError, TypeError]
        nb = NetboxClient(host='http://test', api_key='key')

        with pytest.raises(ValueError):
            await nb.fetch(endpoint='fake', model='valid', params={})

        with pytest.raises(TypeError):
            await nb.fetch(endpoint='valid', model='fake', params={})


@pytest.mark.asyncio
@patch('prophetess_netbox.client.AIONetbox')
async def test_NetboxClient_entity(__aionb):

    with patch.object(NetboxClient, 'fetch', new_callable=asynctest.CoroutineMock) as mf:
        mf.return_value = AIONetboxResponseMock()
        mf.return_value.count = 1
        mf.return_value.results = ['test']

        nb = NetboxClient(host='http://test', api_key='key')

        entity = await nb.entity(
            endpoint='test',
            model='testing',
            params={}
        )

        mf.assert_called_with(
            endpoint='test',
            model='testing',
            params={},
        )

        assert entity == 'test'


@pytest.mark.asyncio
@patch('prophetess_netbox.client.AIONetbox')
async def test_NetboxClient_entity_empty(__aionb):

    with patch.object(NetboxClient, 'fetch', new_callable=asynctest.CoroutineMock) as mf:
        mf.return_value = AIONetboxResponseMock()
        mf.return_value.count = 0

        nb = NetboxClient(host='http://test', api_key='key')

        entity = await nb.entity(
            endpoint='test',
            model='testing',
            params={}
        )

        assert entity is None


@pytest.mark.asyncio
@patch('prophetess_netbox.client.AIONetbox')
async def test_NetboxClient_entity_too_many(__aionb):

    with patch.object(NetboxClient, 'fetch', new_callable=asynctest.CoroutineMock) as mf:
        mf.return_value = AIONetboxResponseMock()
        mf.return_value.count = 2

        nb = NetboxClient(host='http://test', api_key='key')

        with pytest.raises(InvalidPKConfig):
            await nb.entity(
                endpoint='test',
                model='testing',
                params={}
            )


@pytest.mark.asyncio
@patch('prophetess_netbox.client.AIONetbox')
async def test_NetboxClient_entities(__aionb):

    with patch.object(NetboxClient, 'fetch', new_callable=asynctest.CoroutineMock) as mf:
        mf.return_value = AIONetboxResponseMock()
        mf.return_value.count = 1
        mf.return_value.results = ['test']

        nb = NetboxClient(host='http://test', api_key='key')

        entity = await nb.entities(
            endpoint='test',
            model='testing',
            params={}
        )

        mf.assert_called_with(
            endpoint='test',
            model='testing',
            params={},
        )

        assert entity == ['test']


@pytest.mark.asyncio
@patch('prophetess_netbox.client.AIONetbox')
async def test_NetboxClient_entities_empty(__aionb):

    with patch.object(NetboxClient, 'fetch', new_callable=asynctest.CoroutineMock) as mf:
        mf.return_value = AIONetboxResponseMock()
        mf.return_value.count = 0

        nb = NetboxClient(host='http://test', api_key='key')

        entity = await nb.entities(
            endpoint='test',
            model='testing',
            params={}
        )

        assert entity is None



"""Client for Netbox API transactions."""

import asyncio
import logging

from aionetbox import AIONetbox

from prophetess_netbox.exceptions import (
    InvalidPKConfig,
    InvalidNetboxEndpoint,
    InvalidNetboxOperation,
)

log = logging.getLogger('prophetess.plugins.netbox.client')


class NetboxClient:
    """Re-usable abstraction to aionetbox"""

    def __init__(self, *, host, api_key, loop=None):
        """Initialize a single instance with no authentication."""
        self.loop = loop or asyncio.get_event_loop()
        self.__cache = {}  # TODO: make a decorator that caches api classes?

        self.client = AIONetbox.from_openapi(url=host, api_key=api_key)

    async def close(self):
        await self.client.close()

    def build_model(self, endpoint, method, action):
        """ Return the aionetbox Api method from an endpoint class """
        name = '{}_{}_{}'.format(endpoint, method, action)

        try:
            api = getattr(self.client, endpoint)
        except AttributeError:
            raise InvalidNetboxEndpoint('{} module not found'.format(endpoint))

        try:
            return getattr(api, name)
        except AttributeError:
            raise InvalidNetboxOperation('{} not a valid operation'.format(name))

    async def fetch(self, *, endpoint, model, params):
        func = self.build_model(endpoint, model, 'list')
        try:
            return await func(**params)
        except ValueError:
            # Bad Response
            raise
        except TypeError:
            # Bad params
            raise

    async def entity(self, *, endpoint, model, params):
        """ Fetch a single record from netbox using one or more look up params """

        data = await self.fetch(endpoint=endpoint, model=model, params=params)

        if data.count < 1:
            return None

        elif data.count > 1:
            kwargs = ', '.join('='.join(i) for i in params.items())
            raise InvalidPKConfig('Not enough criteria for <{}({})>'.format(endpoint, kwargs))

        return data.results.pop(-1)

    async def entities(self, *, endpoint, model, params):
        """ Fetch all matching records from netbox using one or more look up params """

        data = await self.fetch(endpoint=endpoint, model=model, params=params)

        if data.count < 1:
            return None

        return data.results




from .loader import NetboxLoader

__title__ = 'prophetess-netbox'
__version__ = '0.2.4-2'
__description__ = 'Netbox plugin for Prophetess'
__author__ = 'Marco Ceppi'
__author_email__ = 'marco@vapor.io'
__url__ = 'https://github.com/vapor-ware/prophetess-netbox'
__license__ = 'GNU General Public License v3.0'




from prophetess.exceptions import ServiceError


class NetboxPluginException(ServiceError):
    """Raised when Netbox plugin encounters errors."""
    pass


class InvalidPKConfig(NetboxPluginException):
    """Raised when Netbox PK configuration yields invalid results."""
    pass


class InvalidNetboxEndpoint(NetboxPluginException):
    """Raised when invalid netbox endpoint is provided."""
    pass


class InvalidNetboxOperation(NetboxPluginException):
    """Raised when invalid netbox model or method is provided."""
    pass
